/*****************************************************************************/
/*
                                  search.c

This module contains three separate functionities.

  1. Keyword search for VWcms content (stripped of HTML tags).

  2. Regular expression search for VWcms content (stripped of HTML tags).

  3. Search text for a string and replace all occurances with another.

The first two are intended for the end-user, the second for the manager.


REGEX EVALUATION
----------------
Uses the GNU RX1.5 regular expression package.
Evaluation is done using case-insensitive, EGREP-compatible matching.

A detailed tutorial on regular expression capabilities and usage is well
beyond the scope of this document.  This summary is only to serve as a quick
mnemonic.  VWcms regular expressions support the following set of operators.

  Operator Overview

  Description                    Usage
  -----------                    -----
  Match-self Operator            Ordinary characters
  Match-any-character Operator   .
  Concatenation Operator         Juxtaposition.
  Repetition Operators           *  +  ? {}
  Alternation Operator           |
  List Operators                 [...]  [^...]
  Grouping Operators             (...)
  Back-reference Operator        \digit
  Anchoring Operators            ^  $
  Backslash Operator             Escape meta-character (e.g. \ ^ . $ | [ ()

The following operators are used to match one, or in conjunction with the
repetition operators more, characters of the target string.  These single and
leading characters are reserved meta-characters and must be escaped using a
leading backslash ("\") if required as a literal character in the matching
pattern.

  Matching Operators

  Expression        Purpose
  ----------        -------
  ^                 Match the beginning of the line
  .                 Match any character
  $                 Match the end of the line
  |                 Alternation (or)
  [abc]             Match only a, b or c
  [^abc]            Match anything except a, b and c
  [a-z0-9]          Match any character in the range a to z or 0 to 9
 
Repetition operators control the extent, or number, of whatever the matching
operators match. These are also reserved meta-characters and must be escaped
using a leading backslash if required as a literal character.

  Repetition Operators

  Expression        Function
  ----------        --------
  *                 Match 0 or more times
  +                 Match 1 or more times
  ?                 Match 1 or zero times
  {n}               Match exactly n times
  {n,}              Match at least n times
  {n,m}             Match at least n but not more than m times
 
Regular expression equivalents of these common wildcards

  asterisk ('*')           .*
  question mark ('?')      .
  percentage ('%')         .


COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
04-APR-2010  MGD  support regular expression
24-MAR-2008  MGD  initial
*/

/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <types.h>
#include <time.h>

/* VMS related header files */
#include "descrip.h"
#include "ssdef.h"
#include "stsdef.h"

/* application header file */
#include "cgilib.h"
#include "config.h"
#include "content.h"
#include "data.h"
#include "other.h"
#include "search.h"
#include "regex.h"
#include "tag.h"
#include "template.h"

#define FI_LI "SEARCH", __LINE__

#define RESULT_QUOTE_WORD_COUNT 14

#undef ISLWS
#define ISLWS(ch) (ch == ' ' || ch == '\t')

/* global storage */

/* prototypes */

/* external storage */

extern BOOL  Debug,
             ManageEditOn,
             WatchEnabled;

extern char  ContentFileName[],
             ContentSiteSlashName[],
             SoftwareId[],
             Utility[];

extern CONTENT_DATA  *ContentSitePtr,
                     *ContentSectionPtr;

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
Called by a $$RESULT$$ in the content of a page.

The search keyword is contained in the query-string form field "keyword=".

The function steps through each (generally non-$$) section of the content.  It
allocates sufficient dynamic memory to contain the HTML of the section and then
proceeds to strip all HTML tags from it and compress white-space.  This leaves
what is (fit-for-purpose) just the plain text of the section.

This plain-text is then searched hit-by-hit until exhausted.  Each hit is
provided as output.  When exhausted the allocated memory is freed and then
next section processed in the same manner.  The search ends when the list of
sections is exhausted.
*/

void SearchResult (char *SubsArg)

{
   int  cnt, len,
        HitCount,
        LineCount,
        ExtractCount = RESULT_QUOTE_WORD_COUNT,
        TotalCount;
   char  *cptr, *sptr, *zptr,
         *GotoPtr,
         *KeyWordPtr,
         *SearchPtr,
         *TextPtr,
         *UrlKeyWordPtr;
   char  ExtractBuffer [512],
         StringGoto [256];
   CONTENT_DATA  *cdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SearchResult()\n");

   if (!(KeyWordPtr = GetCgiVarNull ("FORM_KEYWORD")))
      KeyWordPtr = "??KEYWORD??";
   UrlKeyWordPtr = UrlEncode(KeyWordPtr);

   if (*(cptr = SubsArg) == '$') cptr++;

   if (*cptr == '?')
   {
      /* output the keyword used in the search */
      fputs (HtmlEscape(KeyWordPtr), stdout);
      return;
   }

   if (*(USHORTPTR)KeyWordPtr == '\?\?') return;

   if (isdigit(*cptr))
   {
      ExtractCount = atoi(cptr);
      while (*cptr && *cptr != '$') cptr++;
      if (*cptr) cptr++;
   }

   zptr = (sptr = StringGoto) + sizeof(StringGoto)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   if (!*(GotoPtr = StringGoto)) GotoPtr = CONFIG_LANG_PARSE(LangSearch,3);

   if (WatchEnabled) WatchThis (FI_LI, "SEARCH !AZ", KeyWordPtr);

   TotalCount = 0;

   for (cdptr = ContentSitePtr; cdptr; cdptr = cdptr->NextPtr)
   {
      if (!ManageEditOn)
      {
         if (cdptr->IsHidden) continue;
         if (cdptr->GoneToPtr) continue;
      }

      TextPtr = calloc (1, cdptr->TextLength+1);
      if (!TextPtr) ErrorExit (vaxc$errno, FI_LI);

      /* remember, the content is not null-terminated */
      memcpy (TextPtr, cdptr->TextPtr, cdptr->TextLength);
      TextPtr[cdptr->TextLength] = '0';

      SearchStripMarkup (TextPtr);

      /***************/
      /* search text */
      /***************/

      HitCount = 0;
      SearchPtr = TextPtr;
      for (;;)
      {
         if (*KeyWordPtr == '^')
            sptr = SearchRegEx (SearchPtr, KeyWordPtr+1);
         else
            sptr = SearchKeyWord (SearchPtr, KeyWordPtr);
         if (!sptr) break;

         /********/
         /* hit! */
         /********/

         if (!TotalCount++) fputs ("<dl>\n", stdout);

         if (!HitCount++)
            fprintf (stdout,
"<dt class=\"vwcms_result\"><a href=\"%s/%s\">%s</a></dt>\n",
                     UrlEncodeName(ContentSiteSlashName),
                     UrlEncodeName(cdptr->SectionName),
                     HtmlEscape(cdptr->SectionTitle));

         /***************/
         /* get extract */
         /***************/

         /* go back the specified number of 'words' */
         cptr = sptr;
         for (cnt = 0; cnt < ExtractCount/2; cnt++)
         {
            while (cptr > TextPtr && isspace(*cptr)) cptr--;
            if (cptr <= TextPtr) break;
            while (cptr > TextPtr && !isspace(*cptr)) cptr--;
            if (cptr <= TextPtr) break;
         }
         while (isspace(*cptr)) cptr++;
         /* go forward the specified number of 'words' */
         sptr = cptr;
         for (cnt = 0; cnt < ExtractCount; cnt++)
         {
            while (*sptr && isspace(*sptr)) sptr++;
            if (!*sptr) break;
            while (*sptr && !isspace(*sptr)) sptr++;
            if (!*sptr) break;
         }
         if (!*sptr)
         {
            /* ran out of words! get them (if available) going backwards */
            cptr = --sptr;
            for (cnt = 0; cnt < ExtractCount; cnt++)
            {
               while (cptr > TextPtr && isspace(*cptr)) cptr--;
               if (cptr <= TextPtr) break;
               while (cptr > TextPtr && !isspace(*cptr)) cptr--;
               if (cptr <= TextPtr) break;
            }
            while (isspace(*cptr)) cptr++;
         }

         /* copy to a buffer while HTML-escaping */
         len = CgiLibHtmlEscape (cptr, sptr-cptr,
                                 ExtractBuffer, sizeof(ExtractBuffer));

         /* find line number (after copying the extract!) */
         LineCount = 0;
         for (cptr = TextPtr; cptr < sptr; cptr++)
            if (*cptr == '\n') LineCount++;

         /* store the current end of extract for the next (if any) search */
         SearchPtr = sptr;

         /* eliminate multiple white-space from the extract */
         cptr = sptr = ExtractBuffer;
         while (*cptr)
         {
            if (isspace(*cptr))
            {
               *sptr++ = ' ';
               while (isspace(*cptr)) cptr++;
            }
            else
               *sptr++ = *cptr++;
         }
         *sptr = '\0';

         /*******************/
         /* display extract */
         /*******************/

         fprintf (stdout,
"<dd class=\"vwcms_result_quote\">\
<span class=\"vwcms_result_count\">%d.</span>%*.*s\
<span class=\"vwcms_result_goto\">&nbsp;...&nbsp;",
                  TotalCount, len, len, ExtractBuffer);

         if (cptr = IsAnImage (GotoPtr))
            fprintf (stdout,
"<a href=\"%s/%s?result=%s#l%d\">\
<img class=\"vwcms_nav_img\" %s /></a>",
                     UrlEncodeName(ContentSiteSlashName),
                     UrlEncodeName(cdptr->SectionName),
                     UrlKeyWordPtr,
                     LineCount,
                     cptr);
         else
            fprintf (stdout,
"[<a href=\"%s/%s?result=%s#l%d\">%s</a>]",
                     UrlEncodeName(ContentSiteSlashName),
                     UrlEncodeName(cdptr->SectionName),
                     UrlKeyWordPtr,
                     LineCount,
                     GotoPtr);

         fputs ("</span></dd>\n", stdout);
      }

      free (TextPtr);

      if (*KeyWordPtr == '^') SearchRegEx (NULL, NULL);
   }

   if (TotalCount)
      fputs ("</dl>\n", stdout);
   else
      fputs (CONFIG_LANG_PARSE(LangSearch,2), stdout);
}

/*****************************************************************************/
/*
Search for something like "</script>".
Return a pointer to the next character of the end-of-text.
*/

char* SearchEndTag
(
char *FromHere,
char *EndTag
)
{
   int  len;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SearchEndTag()\n");

   len = strlen(EndTag);
   for (cptr = FromHere; *cptr; cptr++)
   {
      while (*cptr && *cptr != '<') cptr++;
      if (!*cptr) break;
      if (strsame (cptr, EndTag, len))
      {
         cptr += len;
         return (cptr);
      }
      cptr++;
   }
   return (cptr);
}

/*****************************************************************************/
/*
Search for the supplied regular expression in the pointed-to text.  The
search is case-insensitive.  If a hit occurs return a pointer to the next
character after the hit.  If a hit does not occur return a NULL.
*/

char* SearchRegEx
(
char *TextPtr,
char *RegExpress
)
{
   static int  RegexCflags = REG_EXTENDED | REG_ICASE,
               RegexEflags = 0;

   static regex_t  *rexptr;
   static regex_t  regexp;

   int  retval;
   char  RegExErrMsg [256];
   regmatch_t  pmatch;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SearchRegEx()\n");

   if (!TextPtr)
   {
      /* initialise */
      if (!rexptr) return (NULL);
      regfree (&regexp);
      rexptr = NULL;
      return (NULL);
   }

   if (!rexptr)
   {
      retval = regcomp (&regexp, RegExpress, RegexCflags);
      if (retval)
      {
         regerror (retval, &regexp, RegExErrMsg, sizeof(RegExErrMsg));
         TemplateAdvise ("Eregex: !AZ", RegExErrMsg);
         regfree (&regexp);
         return (NULL);
      }
      rexptr = &regexp;
   }

   retval = regexec (rexptr, TextPtr, 1, &pmatch, RegexEflags);
   if (retval)
   {
      regfree (&regexp);
      rexptr = NULL;
      return (NULL);
   }

   return (TextPtr + pmatch.rm_so);
}

/*****************************************************************************/
/*
Search for the supplied keyword (or other string) in the pointed-to text.  The
search is case-insensitive.  If a hit occurs return a pointer to the next
character after the hit.  If a hit does not occur return a NULL.
*/

char* SearchKeyWord
(
char *TextPtr,
char *KeyWord
)
{
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SearchKeyWord()\n");

   if (WatchEnabled) WatchThis (FI_LI, "SEARCH |!AZ", KeyWord);

   while (*TextPtr)
   {
      if (tolower(*TextPtr) != tolower(*KeyWord))
      {
         TextPtr++;
         continue;
      }
      cptr = TextPtr;
      sptr = KeyWord;
      for (;;)
      {
         if (!*cptr || !*sptr) break;
         if (tolower(*cptr) != tolower(*sptr)) break;
         cptr++;
         sptr++;
      }
      if (!*sptr) return (cptr);
      TextPtr++;
   }
   return (NULL);
}

/*****************************************************************************/
/*
Simple keyword search and replace.  The search is case-insensitive.  The
replace is done as-is.  The search keyword can include any printable character
except the newline.  The search will not span newlines.   If there was nothing
hit to replace just return a NULL.  Otherwise allocate dynamic memory and
replace each string.   The first line of the (re)built text is a decimal number
containg the number of hits.  This (obviously) should be absorbed by the
calling routime.  Return a pointer to the new text.  The original text is left
untouched regardless.
*/

char* SearchAndReplace
(
char *TextPtr,
char *SearchPtr,
char *ReplacePtr
)
{
   int  CopyLength,
        HitCount,
        NewTextLength,
        ReplaceLength,
        SearchLength,
        TextLength;
   char  *cptr, *sptr,
         *NewTextPtr,
         *SearchTextPtr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled)
      WatchThis (FI_LI, "SEARCH+REPLACE |!AZ|!AZ", SearchPtr, ReplacePtr);

   SearchLength = strlen(SearchPtr);
   ReplaceLength = strlen(ReplacePtr);

   /*************/
   /* how many? */
   /*************/

   HitCount = 0;
   SearchTextPtr = TextPtr;
   while (SearchTextPtr = SearchKeyWord (SearchTextPtr, SearchPtr))
   {
      HitCount++;
      sptr = SearchTextPtr;
   }

   if (WatchEnabled) WatchThis (FI_LI, "HITS !UL", HitCount);

   if (!HitCount) return (NULL);

   /* establish the end-of-text from the last hit (saves some) */
   while (*sptr) sptr++;
   TextLength = sptr - TextPtr;

   /********************/
   /* replace the text */
   /********************/

   NewTextLength = TextLength - (SearchLength * HitCount) +
                                (ReplaceLength * HitCount);

   NewTextPtr = sptr = CgiLibVeeMemCalloc (NewTextLength+64);
   if (!NewTextPtr) ErrorExit (vaxc$errno, FI_LI);

   /* the first line contains the number of hits */
   sptr += sprintf (sptr, "%d\n", HitCount);

   SearchTextPtr = cptr = TextPtr;
   while (SearchTextPtr = SearchKeyWord (SearchTextPtr, SearchPtr))
   {
      CopyLength = SearchTextPtr - cptr - SearchLength;
      memcpy (sptr, cptr, CopyLength);
      cptr = SearchTextPtr;
      sptr += CopyLength;
      memcpy (sptr, ReplacePtr, ReplaceLength);
      sptr += ReplaceLength;
   }
   strcpy (sptr, cptr);

   if (WatchEnabled)
   {
      while (*sptr) sptr++;
      WatchThis (FI_LI, "!UL -> !UL bytes", TextLength, sptr-NewTextPtr);
   }

   return (NewTextPtr);
}

/*****************************************************************************/
/*
Strip VWcms markup, HTML markup and convert HTML entities to real characters,
in situ.  Return the length of the munged string.
*/

int SearchStripMarkup (char *TextPtr)

{
   char  *cptr, *sptr, *tptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SearchStripMarkup()\n");

   cptr = sptr = TextPtr;
   while (*cptr)
   {
      while (*cptr && *cptr != '<' && *(USHORTPTR)cptr != '$$')
         *sptr++ = *cptr++;

      if (!*cptr) break;

      if (*(USHORTPTR)cptr == '$$')
      {
         /****************/
         /* VWcms markup */
         /****************/

         cptr += 2;

         if (*(USHORTPTR)cptr == '$$')
         {
            /* escaped double-quote - we want this in there! */
            *sptr++ = *cptr++;
            *sptr++ = *cptr++;
            continue;
         }

         if (*cptr == '!')
         {
            /* commented-out */
            cptr++;
            while (*cptr)
            {
               while (*cptr && *cptr != '!') cptr++;
               if (*cptr) cptr++;
               if (*(USHORTPTR)cptr != '$$') continue;
               cptr += 2;
               break;
            }
            continue;
         }

         if (strsame (cptr, "SYNOPSIS$text=", 14))
         {
            for (cptr += 14;
                 *cptr && *(USHORTPTR)cptr != '$$';
                 *sptr++ = *cptr++);
            if (!*cptr) break;
            cptr += 2;
            continue;
         }

         if (strsame (cptr, "$$TAG$", 6))
         {
           if (tptr = TagVisible (cptr))
              while (*tptr) *sptr++ = *tptr++;
           /* now drop through to have $$TAG$..$$ absorbed */
         }

         while (*cptr && *(USHORTPTR)cptr != '$$') cptr++;
         if (!*cptr) break;
         cptr += 2;
         continue;
      }

      /*******************/
      /* tag-opening '<' */
      /*******************/

      if (strsame (cptr, "<applet", 7))
      {
         cptr = SearchEndTag (cptr, "</applet>");
         continue;
      }
      if (strsame (cptr, "<head", 5))
      {
         cptr = SearchEndTag (cptr, "</head>");
         continue;
      }
      if (strsame (cptr, "<object", 7))
      {
         cptr = SearchEndTag (cptr, "</object>");
         continue;
      }
      if (strsame (cptr, "<script", 7))
      {
         cptr = SearchEndTag (cptr, "</script>");
         continue;
      }
      if (strsame (cptr, "<style", 6))
      {
         cptr = SearchEndTag (cptr, "</style>");
         continue;
      }

      /*******/
      /* jat */
      /*******/

      /* find the closing bracket of the tag */
      while (*cptr && *cptr != '>') cptr++;
      if (*cptr) cptr++;
   }
   *sptr = '\0';

   /*****************/
   /* HTML entities */
   /*****************/

   ContentDeEntify (TextPtr, FALSE);

   /* compress white-space */
   cptr = sptr = TextPtr;
   while (isspace(*cptr) || (uchar)*cptr == 160) cptr++;
   while (*cptr)
   {
      /* and turn non-breaking-spaces into real spaces */
      if (isspace(*cptr) || (uchar)*cptr == 160)
      {
         *sptr++ = ' ';
         while (isspace(*cptr) || (uchar)*cptr == 160) cptr++;
      }
      else
         *sptr++ = *cptr++;
   }
   while (sptr > TextPtr && (isspace(*(sptr-1)) || (uchar)*(sptr-1) == 160))
      sptr--;
   *sptr = '\0';

   return (sptr - TextPtr);
}

/*****************************************************************************/

